#!/usr/bin/python3
# -*- coding: utf-8 -*-
# @Time    : 2022/5/28 14:09
# @Author  : 马赫
# @Email   : 1692303843@qq.com
# @FileName: train.py


import warnings

warnings.filterwarnings("ignore")

try:
    from BiVFM.func import *
    from BiVFM.olga_func import *
    from BiVFM.model import *
    from BiVFM.feature_engineering import *
    from BiVFM.config_transfer import *
except Exception:
    from func import *
    from olga_func import *
    from model import *
    from feature_engineering import *
    from config_transfer import *

import paddle.nn.functional as F
from paddle.io import DataLoader
from paddle.optimizer import *

# ============== 神经网络部分 ==============


def predict_loader(model, loader):
    """
    输入模型和迭代器，
    如果迭代器同时存在数据和标签，则返回预测的y和真实y，否则只返回预测的y
    y二维矩阵[时间步, 特征维度]
    """
    # TODO 不知道为什么，这样next迭代每次迭代都会增加内存占用

    # # 是否同时存在数据和标签
    # for batch_id, _data in enumerate(loader()):
    #     break

    # if len(_data) == 2:

    # # 只有数据的情况
    # else:
    #     predict_list = []
    #     for batch_id, data in enumerate(loader()):
    #         data = data.astype('float32')
    #         output = model(data)
    #         predict_list.append(output.numpy())
    #     return np.vstack(predict_list)

    predict_list = []
    y_list = []
    for batch_id, (data, label) in enumerate(loader()):
        data = data.astype('float32')
        label = label.astype('float32')
        output = model(data)

        predict_list.append(output.numpy())
        y_list.append(label.numpy())
    return np.vstack(predict_list), np.vstack(y_list)


def visualization(model, loader, standardized_args=None, is_show=False, save_path=None, title=[], show_error=True):
    """
    真实值与预测效果对比可视化

    model：神经网络模型
    loader：数据迭代器
    如果不相等需要将raw_y截取后loader拼接后的长度

    mean, std：反归一化的均值和方差，一维向量
    save_path：保存路径，默认不保存
    """

    # 同时返回y_predict和y_gt
    y_predict, y_gt = predict_loader(model, loader)  # (N, Timesteps, C_out)
    # 将时间滑窗变为时序数据
    y_predict, y_gt = [series2time(y) for y in [y_predict, y_gt]]  # (Timesteps+?, C_out)
    # 将标准化的预测值变回原数据范围
    if standardized_args is not None:
        y_predict, y_gt = [stand2raw(y, standardized_args) for y in [y_predict, y_gt]]
    draw1(gt=y_gt, pred=y_predict, is_show=is_show, img_path=save_path, title=title, show_error=show_error)

    return y_predict, y_gt


class MyIterableDataset(paddle.io.IterableDataset):
    def __init__(self, data_list, x_length, y_length, x_features, y_features, step=1, return_y_info=False):
        """
        输入标准化后的数据
        """
        self.data_list = data_list
        self.x_length = x_length
        self.y_length = y_length
        self.x_features = x_features
        self.y_features = y_features
        self.step = step
        # 是否返回输出y时刻的x信息
        self.return_y_info = return_y_info

    def __iter__(self):

        for data in self.data_list:
            # 返回输出y时刻的x信息
            if self.return_y_info:
                x_window, y_window, y_x_info = generate_xy(data=data,
                                                           x_length=self.x_length, y_length=self.y_length,
                                                           x_features=self.x_features, y_features=self.y_features,
                                                           step=self.step,
                                                           return_y_info=self.return_y_info)
                for x, y, z in zip(x_window, y_window, y_x_info):
                    yield x, y, z

            # 不返回输出y时刻的x信息
            else:
                x_window, y_window = generate_xy(data=data,
                                                 x_length=self.x_length, y_length=self.y_length,
                                                 x_features=self.x_features, y_features=self.y_features,
                                                 step=self.step,
                                                 return_y_info=self.return_y_info)
                for x, y in zip(x_window, y_window):
                    yield x, y


def load_params(model, model_path='save_dir/net.pdparams', opt=None, opt_path='save_dir/opt.pdopts'):
    """
    模型加载
    主要保存模型本身，
    优化器可能有bug建议不读取优化器opt
    """

    assert os.path.isfile(model_path), 'Failed to load model'
    layer_state_dict = paddle.load(model_path)

    model.set_state_dict(layer_state_dict)
    if opt != None:
        assert os.path.isfile(opt_path), 'Failed to load optimizer'
        opt_state_dict = paddle.load(opt_path)
        opt.set_state_dict(opt_state_dict)


def save_params(model, model_path='save_dir/net.pdparams', opt=None, opt_path='save_dir/opt.pdopts'):
    """
    模型保存
    主要保存模型本身，
    优化器可能有bug建议不保存优化器opt
    """

    paddle.save(model.state_dict(), model_path)
    if opt != None:
        paddle.save(opt.state_dict(), opt_path)


class LossFC(paddle.nn.Layer):
    def __init__(self):
        super(LossFC, self).__init__()
        pass

    def forward(self, data, label):
        loss = F.mse_loss(data, label, reduction='mean')
        return loss


def train_model(net, opt, loss_fc,
                num_epoch, continue_epoch,
                train_loader, test_loader,
                scalar_s2r_y,
                y_feature,
                base_conf, feature_list,
                is_show=True
                ):
    """
    net:模型
    opt:模型
    loss_fc:损失函数

    num_epoch: int 总轮次
    continue_epoch: int 从第几轮次继续

    train_loader:训练迭代器
    test_loader:验证绘图迭代器

    scalar_s2r_y: 元组 (全部y的均值向量，全部y的方差向量) 将标准化的y转化为原始尺度

    y_feature: list 矩阵的哪几列是y，一般为最后4列

    base_conf: dict 配置
    feature_list: np.array 全部列名向量

    is_show: 是否显示


    """

    min_loss = 1e9  # 初始化 最小误差
    for epoch in range(continue_epoch, num_epoch):
        net.train()
        loss_list = []
        for batch_id, (data, label) in enumerate(train_loader()):
            data = data.astype('float32')
            label = label.astype('float32')
            output = net(data)

            step_loss = loss_fc(output, label)

            loss_list.append(step_loss.numpy())
            step_loss.backward()
            opt.step()
            opt.clear_grad()
            if batch_id % 100 == 0:
                # 每n轮训练保存一次日志
                loss = sum(loss_list) / len(loss_list)
                log("Training epoch: %s batch_id: %s, loss: %s" % (epoch, batch_id, loss),
                    log_file=base_conf['train_log_file'])

        loss = sum(loss_list) / len(loss_list)

        if loss <= min_loss:
            # 当前损失值小于历史最低的损失值时，进行测试
            # 保存模型，并可视化，记录日志

            # 将模型转为测试模式
            net.eval()
            # 保存模型的网络结构参数
            save_params(net, model_path=base_conf['model_path'])

            # 获取训练过程的图像保存路径
            save_path = os.path.join(base_conf['save_path'], '%s.jpg' % epoch)
            # 训练过程的训练效果可视化
            visualization(model=net, loader=test_loader,
                          standardized_args=scalar_s2r_y,
                          is_show=is_show, save_path=save_path,
                          title=feature_list[y_feature],
                          show_error=False)

            # 将每次获得最佳成绩的记录保存日志
            content = "Current Epoch: %s, loss %s  better than last loss %s, save model" % (epoch, loss, min_loss)
            log(content, log_file=base_conf['train_log_file'])
            min_loss = loss

        if epoch % 1 == 0:
            # 每n轮训练保存一次日志
            log("Training epoch: %s, loss: %s" % (epoch, loss), log_file=base_conf['train_log_file'])

    log("Training till end", log_file=base_conf['train_log_file'])


def train_process(base_yaml_path):
    """
    One button to start training
    :param base_yaml_path:str, config file absolute path
    :return: None
    """

    base_conf = get_base_config(base_yaml_path)
    # print(base_conf)

    pkl_dir = base_conf['pkl_dir']
    num_epoch = base_conf['num_epoch']
    input_channels = base_conf['input_channels']
    output_channels = base_conf['output_channels']

    continue_epoch = base_conf['continue_epoch']  # 从多少轮开始训练

    ratio = base_conf['ratio']
    model_name = base_conf['model_name']

    load_model = base_conf['load_model']

    # 加载数据
    raw_data_list = pickle_load('%s/raw_data_list.pkl' % pkl_dir)
    print('Number of sample files:', len(raw_data_list))

    # 以全部文件作为标准化基准
    all_mean = []
    all_std = []
    for i in raw_data_list:
        mean = i.mean()
        std = i.std()
        all_mean.append(mean)
        all_std.append(std)

    all_mean = pd.concat(all_mean, axis=1).T
    all_std = pd.concat(all_std, axis=1).T

    standardized_args = (all_mean.mean(), all_std.mean())

    stand_data_list = [raw2stand(raw_data, standardized_args) for raw_data in raw_data_list]

    # 切分特征
    feature_list = np.array(raw_data_list[0].columns)

    # print('Dataset Characteristics:')
    # for num, feature in enumerate(feature_list):
    #     print(num, feature)
    # print('=' * 10)

    x_feature = list(range(len(feature_list)))[:-output_channels]
    y_feature = list(range(len(feature_list)))[-output_channels:]

    print('x特征%s列，y特征%s列' % (len(x_feature), len(y_feature)))
    print('========== x_feature: ==========\n%s' % feature_list[x_feature])
    print('========== y_feature: ==========\n%s' % feature_list[y_feature])

    np.random.seed(0)
    np.random.shuffle(stand_data_list)

    num_ratio = int(len(stand_data_list) * ratio)
    train_data_list = stand_data_list[: num_ratio]
    test_data_list = stand_data_list[num_ratio:]

    # DataLoader
    train_reader = MyIterableDataset(train_data_list, base_conf['x_length'], base_conf['y_length'], x_feature,
                                     y_feature)

    test_reader = MyIterableDataset(test_data_list[: 10], base_conf['x_length'], base_conf['y_length'],
                                    x_feature,
                                    y_feature, return_y_info=False)

    train_loader = DataLoader(train_reader, batch_size=base_conf['batch_size'], drop_last=False)
    test_loader = DataLoader(test_reader, batch_size=base_conf['batch_size'], drop_last=False)

    for batch_id, (data, label) in enumerate(train_loader()):
        print('train batch x, y shape\t', data.shape, label.shape)
        break

    # scalar_s2r = None
    # scalar_s2r_x = (standardized_args[0].iloc[x_feature].values, standardized_args[1].iloc[x_feature].values)
    scalar_s2r_y = (standardized_args[0].iloc[y_feature].values, standardized_args[1].iloc[y_feature].values)

    net = Net(input_channels, output_channels, time_steps=base_conf['y_length'], model=model_name)

    opt = Adam(learning_rate=base_conf['learning_rate'], parameters=net.parameters())  # Adam

    loss_fc = LossFC()

    # paddle.summary(Net(input_channels, output_channels), (128, 187, 5))

    # a = Net(input_channels, time_steps=1)
    # shape = (128, 187, 35)
    # print('input shape', shape)
    # print('output shape', a(paddle.randn(shape)).shape)
    # paddle.summary(a, shape)

    if load_model:
        try:
            load_params(net, model_path=base_conf['model_path'])
            log('模型已加载', log_file=base_conf['train_log_file'])
        except Exception as e:
            log('未加载模型', log_file=base_conf['train_log_file'])

    else:
        log('未加载模型', log_file=base_conf['train_log_file'])

    train_model(net=net, opt=opt, loss_fc=loss_fc, num_epoch=num_epoch, continue_epoch=continue_epoch,
                train_loader=train_loader, test_loader=test_loader, scalar_s2r_y=scalar_s2r_y, y_feature=y_feature,
                base_conf=base_conf, feature_list=feature_list, is_show=True)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# @Time    : 2022/6/7 11:46
# @Author  : 马赫
# @Email   : 1692303843@qq.com
# @FileName: config_transfer.py


try:
    from BiVFM.func import *
except Exception:
    from func import *


def get_base_config(base_yaml_path):
    """
    get config from yaml and update

    base_yaml_path = r'xxx\demo_work1\conf_0_base.yaml'
    base_conf = get_base_config(base_yaml_path)
    for k, v in base_conf.items():
        print('%s:\t%s' % (k, v))

    :param base_yaml_path: str config file path
    :return: dict of config
    """

    def path_join(*args):
        return '/'.join(args)

    base_conf = yaml_read(base_yaml_path)

    # =============== base config update ===============
    base_conf['learning_rate'] = float(base_conf['learning_rate'])


    base_conf['autoolga_output_dir'] = path_join(base_conf['base_dir'], base_conf['autoolga_output_dir'])

    base_conf['extract_xy'] = path_join(base_conf['base_dir'], base_conf['extract_xy'])

    base_conf['tpl_dir'] = path_join(base_conf['base_dir'], base_conf['tpl_dir'])
    base_conf['raw_csv_dir'] = path_join(base_conf['base_dir'], base_conf['raw_csv_dir'])
    base_conf['xy_dir'] = path_join(base_conf['base_dir'], base_conf['xy_dir'])
    base_conf['pkl_dir'] = path_join(base_conf['base_dir'], base_conf['pkl_dir'])

    base_conf['mean_path'] = path_join(base_conf['base_dir'], base_conf['mean_path'])
    base_conf['std_path'] = path_join(base_conf['base_dir'], base_conf['std_path'])


    base_conf['train_log_file'] = path_join(base_conf['base_dir'], base_conf['info_path'], base_conf['train_log_file'])
    base_conf['other_log_file'] = path_join(base_conf['base_dir'], base_conf['info_path'], base_conf['other_log_file'])

    base_conf['save_path'] = path_join(base_conf['base_dir'], base_conf['info_path'], base_conf['save_path'])
    base_conf['img_save_path'] = path_join(base_conf['base_dir'], base_conf['info_path'], base_conf['img_save_path'])

    base_conf['opt_path'] = path_join(base_conf['base_dir'], base_conf['info_path'], base_conf['opt_path'])
    base_conf['model_path'] = path_join(base_conf['base_dir'], base_conf['info_path'], base_conf['model_path'])

    return base_conf


if __name__ == '__main__':
    base_yaml_path = r'D:\work\VFM\BiVFM\program\demo_work1\conf_0_base.yaml'
    base_conf = get_base_config(base_yaml_path)
    for k, v in base_conf.items():
        print('%s\t%s\t%s' % (k, type(v), v))






#!/usr/bin/python3
# -*- coding: utf-8 -*-
# @Time    : 2022/5/28 11:30
# @Author  : 马赫
# @Email   : 1692303843@qq.com
# @FileName: func.py


import os
import shutil
import warnings

import numpy as np
import pandas as pd

warnings.filterwarnings("ignore")

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib import rcParams

plt_config = {
    "font.family": 'Times New Roman',  # 设置字体类型
    "font.size": 13,
}
rcParams.update(plt_config)

import yaml
import pickle
import logging


# ============== 公共方法 ==============


def rm_dir(path):
    """
    保证该文件夹不存在
    """
    if os.path.isdir(path):
        shutil.rmtree(path)


def remove_list(base='olga_work', removes=['log_imgs', 'save_dir', 'imgs_save', 'logs']):
    """
    删除列表中的文件或文件夹
    """
    for i in removes:
        try:
            abs_i = '%s/%s' % (base, i)
            shutil.rmtree(abs_i)
        except Exception as e:
            pass


def log(content, log_file='logs/log.txt'):
    """
    logging库，将content内容同时输出到控制台，并追加到log_file
    """

    # 确保存在该路径
    path = os.path.split(log_file)[0]
    if not os.path.isdir(path):
        os.makedirs(path)

    logger = logging.getLogger()
    # 将logging模块的handlers列表清空
    # 这行代码很重要！！
    logger.handlers = []

    logger.setLevel(logging.INFO)

    formatter = logging.Formatter("[%(asctime)s] %(filename)s [line:%(lineno)d] %(levelname)s: %(message)s")

    fh = logging.FileHandler(log_file, mode='a')
    fh.setLevel(logging.DEBUG)
    fh.setFormatter(formatter)

    sh = logging.StreamHandler()
    sh.setFormatter(formatter)

    logger.addHandler(sh)
    logger.addHandler(fh)

    logger.info(content)


def makedirs(path):
    """
    保证path这个文件或文件夹路径存在，如果没有则创建该路径
    path 可以为文件或文件夹
    """
    # 通过是否有.xxx后缀判断是文件或文件夹
    # 如果是文件则去除文件名
    # 如果是文件夹则直接判断该文件夹是否存在并创建
    if os.path.splitext(path)[1] != '':
        # 这个是文件
        path = os.path.split(path)[0]

    if not os.path.isdir(path):
        os.makedirs(path)


def pickle_save(path, content):
    """
    保存py变量
    """
    # 如果不存在该路径，则创建文件夹
    makedirs(path)
    with open(path, 'wb') as f:
        pickle.dump(content, f)


def pickle_load(path):
    """
    读取py变量
    """
    with open(path, 'rb') as f:
        content = pickle.load(f)
    return content


def read_df_second_resample(filename, resample_time='s'):
    """
    从csv文件读取数据，并以每秒1条记录进行插值
    第一列为时间索引，间隔可能大于或小于1秒，将其重采样到1秒
    """
    _raw_data = pd.read_csv(filename, header=0,
                            index_col=0,  # 第一列读取为时间索引
                            parse_dates=[0],
                            date_parser=df_parser
                            )

    # 原始数据时间不均匀，重采样到每秒1个点
    # raw_data = _raw_data.resample('S').mean()
    raw_data = _raw_data.resample(resample_time).mean()
    # 重采样后的缺失值进行线性插值
    raw_data = raw_data.interpolate(method='linear')
    print('原始数据形状：%s，重采样后形状：%s，文件名：%s' % (_raw_data.shape, raw_data.shape, filename))

    return raw_data


# ============== 数据处理部分 ==============


def yaml_read(yaml_path):
    """
    yaml配置读取器，
    输入yaml文件路径
    返回变量与值的字典
    """
    try:
        with open(yaml_path) as f:
            _info_yaml = f.read()
    except UnicodeDecodeError:
        # print('捕获编码问题，尝试中文编码')
        with open(yaml_path, encoding='utf-8') as f:
            _info_yaml = f.read()

    info_yaml = yaml.safe_load(_info_yaml)
    return info_yaml


def df_parser(x, mode=0, unit='s', date_format='%Y-%m-%d %H:%M:%S'):
    """
    pd.DataFrame读取时，时间解析格式
    pd.read_csv(data_filename, header=0,
                        index_col=0,  # 第一列读取为索引
                        parse_dates=[0],  # 第一列解析为时间
                        date_parser=df_parser  # 时间解析方法
                        )

    x为该行数据
    mode为解析格式，默认为自适应解析，先解析为以每秒为间隔，如果有问题再次尝试为以日期为解析格式

    """
    if mode == 0:
        # 模式0，将时间字符串先尝试解析为以秒为单位，后尝试解析为日期格式
        try:
            # 先尝试解析为以秒为单位
            z = pd.to_datetime(x, unit=unit)
        except Exception as e:
            try:
                # 后尝试解析为日期格式
                z = pd.to_datetime(x, format=date_format)
            except Exception as e:
                raise e


    elif mode == 1:
        # 模式1，将时间字符串解析为以秒为单位
        z = pd.to_datetime(x, unit=unit)
    elif mode == 2:
        # 模式2，将日期格式解析成指定格式
        z = pd.to_datetime(x, format=date_format)

    return z


def df2list(data, N, length, features, shift, step):
    """
    N = 滑窗个数 需要计算
    length = 滑窗长度 72
    features = 选择特征列 [0, 1, ...]
    shift = 向后偏移行数 length
    step = 滑动间隔 1
    return dy_list：返回滑窗 三维矩阵 [滑窗的个数，每个滑窗的长度，特征维度]
    """
    dy_list = []
    for i in range(N):
        dy = data[i * step + shift: i * step + shift + length, features]
        assert dy.shape[0] == length, '形状%s的长度小于length:%s' % (str(dy.shape), length)
        dy_list.append(dy)

    dy_list = np.array(dy_list)
    # print(dy_list.shape)
    return dy_list


def generate_xy(data, x_length=4, y_length=4, x_features=[0, 1], y_features=[2, 3], step=1, return_y_info=False):
    """
    将二维时序数据转为三维时间滑窗
    data：时序数据 二维矩阵 [时间，特征维度]
    x_length：输入样本和标签的长度，即输入m个时间点的数据
    y_length：输出样本和标签长度，即输出预测未来n个时间点的数据，int, >1
    x_features：特征维度中，样本列号，例如前2列是样本[0, 1]
    y_features：特征维度中，标签列号，例如后3列是标签[-3, -2, -1]
    step：每次滑窗滑动的步长，一般为1，int，>=1
    """

    y_shift = x_length
    x_shift = 0
    N = (len(data) - y_length - y_shift) // step + 1
    # step 余数 需要舍弃，可以舍弃左面或舍弃右面，并打印提示余数，默认舍弃左面
    remainder = (len(data) - y_length - y_shift) % step
    if remainder != 0:
        print('step取 {}时，存在余数remainder {}'.format(step, remainder))

    x = df2list(data=np.array(data), N=N, length=x_length, features=x_features, shift=x_shift + remainder, step=step)
    y = df2list(data=np.array(data), N=N, length=y_length, features=y_features, shift=y_shift + remainder, step=step)

    # 部分算法需要返回预测y窗口对应的x，这个x与输入的x不同，是输出y时刻将记录得到的x
    if return_y_info:
        y_x_info = df2list(data=np.array(data), N=N, length=y_length, features=x_features, shift=y_shift + remainder,
                           step=step)
        return x, y, y_x_info

    return x, y


def period_generate_xy(data_list, x_length, y_length, x_features, y_features, step=1):
    """
    流动周期列表的时间滑窗生成

    """
    x_list, y_list = [], []
    for data in data_list:
        x, y = generate_xy(data.values, x_length, y_length, x_features, y_features, step)

        x_list.append(x)
        y_list.append(y)

    x_list = np.vstack(x_list)
    y_list = np.vstack(y_list)

    return x_list, y_list


# ================================= 标准化方法 =================================

# 均值方差标准化
def raw2stand(raw_data, standardized_args, epsilon=1e-6):
    mean, std = standardized_args
    return (raw_data - mean) / (std + epsilon)


def stand2raw(stand_data, standardized_args, epsilon=1e-6):
    mean, std = standardized_args
    return stand_data * (std + epsilon) + mean


# EVN标准化方法
# def raw2stand(raw_data, mean, std, epsilon=1e-8):
#     return (raw_data - mean) / (mean + epsilon)

# def stand2raw(stand_data, mean, std, epsilon=1e-8):
#     return stand_data * (mean + epsilon) + mean


# 截断truncation log标准化
# def raw2stand(raw_data, *args):
#     new_raw_data = raw_data.copy()
#     _min = min(new_raw_data.min())
#     # assert _min > -100, '最小值超出正常情况:%s' % _min
#     if _min < -100:
#         print('最小值超出正常情况:%s' % _min)
#     new_raw_data[new_raw_data < 0] = 0
#     stand = np.log10(new_raw_data + 10)
#     return stand

# def stand2raw(stand_data, *args):
#     stand = np.power(10, stand_data) - 10
#     return stand


# ================================= 标准化方法 =================================


def series2time(series_tensor, step=1):
    """
    将series_tensor[n个时间窗口, time_step, feature]的时序数据三维张量
    变回为time_matrix[n+time_step-step, feature]原始数据长度二维矩阵

    step：构建时间滑窗时的滑动步长，一般为1
    """
    # 将第一个窗口数据完全保留
    # 从第二个开始的窗口数据只保留后step个
    # 如果step == time_step，则代表窗口每次移动了整个窗口大小，则全部保留
    added_matrix = np.vstack(series_tensor[1:, -step:, :])
    time_matrix = np.vstack([series_tensor[0], added_matrix])
    return time_matrix


def interrupt_time2period_list(raw_data, delta_time="1 hours"):
    """
    原始数据不存在缺失值，大多数时间点之间连续，
    部分时间段出现长时间的跳过，少部分时间点的间隔出现不规律
    每一段连续存在数据的部分作为一个周期截取下来单独保存

    # TODO 增加缺失值的处理功能，比如删除或插值

    raw_data：原始存在时间断层的数据
    delta_time：最长时间间隔，默认1小时
    return period_resample_raw_data_list 将不同流动周期数据拆分成长度不同的元素，
            再组成列表，每个元素都是二维矩阵[时间，特征维度]
    """

    # 连续时间索引
    time_index_continuous = raw_data.index
    # 连续时间索引进行差分
    time_index_continuous_diff = np.ediff1d(time_index_continuous)

    # 每个流动周期的最长时间间隔为1小时
    # 即超过1小时的时间空白认为不同的流动周期
    # 小于一小时则认为是同一个流动周期

    # 每两个时间点的差大于一小时的位置为流动周期 delta_time = "1 hours"
    interval_flag = time_index_continuous_diff > pd.Timedelta(delta_time)
    # 流动周期数据量
    num_period = interval_flag.sum()
    # 流动周期行号
    period_row_list = np.where(interval_flag)[0]

    # 将原始数据分为流动周期的原始数据 period_raw_data_list
    period_raw_data_list = []
    n0 = 0
    n1 = period_row_list[0]
    for n in range(num_period + 1):
        # 从原始数据中截取到每个流动周期的数据
        period_raw_data = raw_data[n0 + 1: n1 + 1]
        period_raw_data_list.append(period_raw_data)

        n0 = n1
        if (n + 1) >= num_period:
            n1 = len(time_index_continuous_diff)
            continue
        n1 = period_row_list[n + 1]

    period_resample_raw_data_list = []
    for period_raw_data in period_raw_data_list:
        period_resample_raw_data = period_raw_data.resample('T').mean()
        period_resample_raw_data = period_resample_raw_data.interpolate(method='linear')

        period_resample_raw_data_list.append(period_resample_raw_data)

    return period_resample_raw_data_list


# ============== matplotlib画图 ==============


def draw1(gt, pred, is_show=False, img_path=None, title=[], show_error=True):
    """
    gt 真实值
    pred 预测值

    title：每列数据的标题名
    """
    fig = plt.figure(figsize=(15, 3))
    for i in range(gt.shape[1]):
        plt.subplot(1, gt.shape[1], i + 1)
        plt.plot(gt[:, i], label='Truth')
        plt.plot(pred[:, i], label='Predict')
        if show_error:
            plt.plot(gt[:, i] - pred[:, i], label='Error', c='g')
        if len(title) == gt.shape[1]:
            temp_title = list(title[i])
            if len(temp_title) > 10:
                # temp_title.insert(10, '\n')
                temp_title = temp_title[:20]
            temp_title = ''.join(temp_title)
            plt.title(temp_title)
        # plt.legend()

    plt.tight_layout()

    if img_path != None:
        makedirs(img_path)
        plt.savefig(img_path, bbox_inches='tight', dpi=300)
    if is_show:
        plt.show()
    else:
        plt.close(fig)


def draw_PT_Qowg(PT_Qowg_data):
    """
    此函数专门用于绘制压力、温度、油气水流量，即二维矩阵[时间点，特征]
    PT_Qowg_data:即二维矩阵

    暂时没有返回和保存功能
    """

    index_plt = PT_Qowg_data.index

    # 画输入 PT 压力温度 图像
    fig, ax1 = plt.subplots()  # figsize=(10, 6)
    ax1.scatter(index_plt, PT_Qowg_data.iloc[:, 0], c='g', s=1)
    plt.xticks(rotation=30)
    plt.legend(loc='upper left')

    ax2 = ax1.twinx()
    ax2.scatter(index_plt, PT_Qowg_data.iloc[:, 1], c='r', s=1)

    plt.xlim([index_plt[0], index_plt[-1]])
    # fig.legend(loc=(0.12, 0.8))
    plt.legend(loc='upper right')

    plt.show()

    # 画输出 Qo Qw Qg  油气水流量 图像
    fig, ax1 = plt.subplots()  # figsize=(10, 6)
    ax1.scatter(index_plt, PT_Qowg_data.iloc[:, 2], c='g', s=1)
    ax1.scatter(index_plt, PT_Qowg_data.iloc[:, 3], c='b', s=1)
    plt.xticks(rotation=30)
    plt.legend(loc='upper left')

    ax2 = ax1.twinx()
    ax2.scatter(index_plt, PT_Qowg_data.iloc[:, 4], c='r', s=1)

    plt.xlim([index_plt[0], index_plt[-1]])
    # fig.legend(loc=(0.09, 0.75))
    plt.legend(loc='upper right')

    plt.show()



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# @Time    : 2022/5/28 14:03
# @Author  : 马赫
# @Email   : 1692303843@qq.com
# @FileName: model.py


import warnings

warnings.filterwarnings("ignore")

import paddle
from paddle import nn
from paddle.nn.utils import weight_norm


# import paddlenlp as nlp


# 单LSTM网络
class LSTM(paddle.nn.Layer):
    def __init__(self, input_channels=7, output_channels=2):
        super(LSTM, self).__init__()
        self.lstm = paddle.nn.LSTM(input_size=input_channels,
                                   hidden_size=50,
                                   num_layers=4,
                                   dropout=0.2,
                                   time_major=False)  # 要求输入的形状是[batch_size,time_steps,input_size]
        self.fc = paddle.nn.Linear(in_features=50, out_features=output_channels)

    def forward(self, inputs):
        outputs, final_states = self.lstm(inputs)  # 使用最后一层的最后一个step的输出作为线性层的输入
        y = self.fc(final_states[0][3])  # 输入：形状为 [batch_size,∗,in_features] 的多维Tensor。
        y = paddle.unsqueeze(y, [1])
        return y


# paddle.summary(LSTM(11, 4), (256, 187, 11))


# 三层LSTM模型，效果差
class TriLSTM(nn.Layer):
    def __init__(self, input_channels=7, output_channels=2):
        super(TriLSTM, self).__init__()

        self.lstm = nn.LSTM(input_size=input_channels, hidden_size=128, num_layers=1)
        self.lstm2 = nn.LSTM(input_size=128, hidden_size=128, num_layers=1)
        self.lstm3 = nn.LSTM(input_size=128, hidden_size=output_channels, num_layers=1)
        self.fc = nn.Linear(187, 1)

    def forward(self, x):
        x, _ = self.lstm(x)
        x, _ = self.lstm2(x)
        x, _ = self.lstm3(x)
        x = paddle.transpose(x, (0, 2, 1))
        x = self.fc(x)
        x = paddle.transpose(x, (0, 2, 1))
        # print(x.shape)

        return x


# print('=' * 40)
# paddle.summary(TriLSTM(11, 4), (256, 187, 11))


# TCN模型
class Chomp1d(nn.Layer):
    """
    Remove the elements on the right.
    Args:
        chomp_size (int): The number of elements removed.
    """

    def __init__(self, chomp_size):
        super(Chomp1d, self).__init__()
        self.chomp_size = chomp_size

    def forward(self, x):
        return x[:, :, :-self.chomp_size]


class TemporalBlock(nn.Layer):
    """
    The TCN block, consists of dilated causal conv, relu and residual block.
    See the Figure 1(b) in https://arxiv.org/pdf/1803.01271.pdf for more details.
    Args:
        n_inputs ([int]): The number of channels in the input tensor.
        n_outputs ([int]): The number of filters.
        kernel_size ([int]): The filter size.
        stride ([int]): The stride size.
        dilation ([int]): The dilation size.
        padding ([int]): The size of zeros to be padded.
        dropout (float, optional): Probability of dropout the units. Defaults to 0.2.
    """

    def __init__(self,
                 n_inputs,
                 n_outputs,
                 kernel_size,
                 stride,
                 dilation,
                 padding,
                 dropout=0.2):
        super(TemporalBlock, self).__init__()
        self.conv1 = weight_norm(
            nn.Conv1D(
                n_inputs,
                n_outputs,
                kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation))
        # Chomp1d is used to make sure the network is causal.
        # We pad by (k-1)*d on the two sides of the input for convolution,
        # and then use Chomp1d to remove the (k-1)*d output elements on the right.
        self.chomp1 = Chomp1d(padding)
        self.relu1 = nn.ReLU()
        self.dropout1 = nn.Dropout(dropout)

        self.conv2 = weight_norm(
            nn.Conv1D(
                n_outputs,
                n_outputs,
                kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation))
        self.chomp2 = Chomp1d(padding)
        self.relu2 = nn.ReLU()
        self.dropout2 = nn.Dropout(dropout)

        self.net = nn.Sequential(self.conv1, self.chomp1, self.relu1,
                                 self.dropout1, self.conv2, self.chomp2,
                                 self.relu2, self.dropout2)
        self.downsample = nn.Conv1D(n_inputs, n_outputs,
                                    1) if n_inputs != n_outputs else None
        self.relu = nn.ReLU()
        self.init_weights()

    def init_weights(self):
        self.conv1.weight.set_value(
            paddle.tensor.normal(0.0, 0.01, self.conv1.weight.shape))
        self.conv2.weight.set_value(
            paddle.tensor.normal(0.0, 0.01, self.conv2.weight.shape))
        if self.downsample is not None:
            self.downsample.weight.set_value(
                paddle.tensor.normal(0.0, 0.01, self.downsample.weight.shape))

    def forward(self, x):
        out = self.net(x)
        res = x if self.downsample is None else self.downsample(x)
        return self.relu(out + res)


class TCNEncoder(nn.Layer):
    """
    A `TCNEncoder` takes as input a sequence of vectors and returns a
    single vector, which is the last one time step in the feature map.
    The input to this encoder is of shape `(batch_size, num_tokens, input_size)`,
    and the output is of shape `(batch_size, num_channels[-1])` with a receptive
    filed:

    .. math::

        receptive filed = 2 * \sum_{i=0}^{len(num\_channels)-1}2^i(kernel\_size-1).

    Temporal Convolutional Networks is a simple convolutional architecture. It outperforms canonical recurrent networks
    such as LSTMs in many tasks. See https://arxiv.org/pdf/1803.01271.pdf for more details.
    Args:
        input_size (int): The number of expected features in the input (the last dimension).
        num_channels (list): The number of channels in different layer.
        kernel_size (int): The kernel size. Defaults to 2.
        dropout (float): The dropout probability. Defaults to 0.2.
    """

    def __init__(self, input_size, num_channels, time_steps=1, kernel_size=2, dropout=0.2):
        super(TCNEncoder, self).__init__()
        self._input_size = input_size
        self._output_dim = num_channels[-1]
        self.time_steps = time_steps
        layers = nn.LayerList()
        num_levels = len(num_channels)
        for i in range(num_levels):
            dilation_size = 2 ** i
            in_channels = input_size if i == 0 else num_channels[i - 1]
            out_channels = num_channels[i]
            layers.append(
                TemporalBlock(
                    in_channels,
                    out_channels,
                    kernel_size,
                    stride=1,
                    dilation=dilation_size,
                    padding=(kernel_size - 1) * dilation_size,
                    dropout=dropout))

        self.network = nn.Sequential(*layers)

    def get_input_dim(self):
        """
        Returns the dimension of the vector input for each element in the sequence input
        to a `TCNEncoder`. This is not the shape of the input tensor, but the
        last element of that shape.
        """
        return self._input_size

    def get_output_dim(self):
        """
        Returns the dimension of the final vector output by this `TCNEncoder`.  This is not
        the shape of the returned tensor, but the last element of that shape.
        """
        return self._output_dim

    def forward(self, inputs):
        """
        TCNEncoder takes as input a sequence of vectors and returns a
        single vector, which is the last one time step in the feature map.
        The input to this encoder is of shape `(batch_size, num_tokens, input_size)`,
        and the output is of shape `(batch_size, num_channels[-1])` with a receptive
        filed:

        .. math::

            receptive filed = 2 * \sum_{i=0}^{len(num\_channels)-1}2^i(kernel\_size-1).
        Args:
            inputs (Tensor): The input tensor with shape `[batch_size, num_tokens, input_size]`.
        Returns:
            Tensor: Returns tensor `output` with shape `[batch_size, num_channels[-1]]`.
        """
        inputs_t = inputs.transpose([0, 2, 1])
        # output = self.network(inputs_t).transpose([2, 0, 1])[-1]
        output1 = self.network(inputs_t)
        output2 = output1.transpose([0, 2, 1])
        output = output2[:, -self.time_steps:, :]

        return output


class TimeSeriesNetwork(nn.Layer):
    def __init__(self, in_features, out_features=1, time_steps=1, num_channels=[256]):
        super(TimeSeriesNetwork, self).__init__()
        self.last_num_channel = num_channels[-1]
        self.tcn = TCNEncoder(in_features, num_channels, time_steps,
                              kernel_size=3, dropout=0.2)
        self.linear = nn.Linear(in_features=self.last_num_channel, out_features=out_features)

    def forward(self, x):
        tcn_out = self.tcn(x)
        y_pred = self.linear(tcn_out)
        return y_pred


# TCNEncoder 输入形状(batch_size, time_steps, features) 输出形状(batch_size, ?)
# paddle.summary(TimeSeriesNetwork(11, 4, time_steps=1, num_channels=[256, ]), (32, 187, 11))


class MFTcNeXt(nn.Layer):
    def __init__(self, in_features, out_features=1, time_steps=1, num_channels=[256], hidden_channels=32):
        super(MFTcNeXt, self).__init__()

        self.tcn1 = TCNEncoder(in_features, num_channels, time_steps, kernel_size=3, dropout=0.2)
        # self.conv1 = nn.Conv1D(num_channels[0], out_features, kernel_size=3, stride=1, padding=1)
        self.fc1 = nn.Linear(num_channels[0], out_features)

        self.tcn2 = TCNEncoder(num_channels[0], num_channels, time_steps, kernel_size=3, dropout=0.2)
        # self.conv2 = nn.Conv1D(num_channels[0], out_features, kernel_size=3, stride=1, padding=1)
        self.fc2 = nn.Linear(num_channels[0], out_features)

        self.tcn3 = TCNEncoder(num_channels[0], num_channels, time_steps, kernel_size=3, dropout=0.2)
        # self.conv3 = nn.Conv1D(num_channels[0], out_features, kernel_size=3, stride=1, padding=1)
        self.fc3 = nn.Linear(num_channels[0], out_features)

        # self.conv = nn.Conv1D(3*time_steps, time_steps, kernel_size=3, stride=1, padding=1)
        self.fc = nn.Linear(3 * time_steps, time_steps)

    def Time_Decoder(self, layer, x):
        x = x.transpose([0, 2, 1])
        x = layer(x)
        x = x.transpose([0, 2, 1])
        return x

    def forward(self, x):
        tcn_out1 = self.tcn1(x)
        # t1_dec = self.Time_Decoder(self.conv1, tcn_out1)
        t1_dec = self.fc1(tcn_out1)

        tcn_out2 = self.tcn2(tcn_out1)
        # t2_dec = self.Time_Decoder(self.conv2, tcn_out2)
        t2_dec = self.fc2(tcn_out2)

        tcn_out3 = self.tcn3(tcn_out2)
        # t3_dec = self.Time_Decoder(self.conv3, tcn_out3)
        t3_dec = self.fc2(tcn_out3)

        # 对时间维度堆叠合并，并一维卷积
        t_dec = paddle.concat([t1_dec, t2_dec, t3_dec], axis=1)
        # y_pred = self.conv(t_dec)
        y_pred = self.Time_Decoder(self.fc, t_dec)

        return y_pred


class PositionalEncoding(nn.Layer):
    def __init__(self, d_model, max_len=5000):
        super(PositionalEncoding, self).__init__()

        pe = paddle.zeros([max_len, d_model])
        position = paddle.arange(0, max_len, dtype='float32').unsqueeze(1)

        div_term = paddle.arange(0, d_model, 2, dtype='float32')
        div_term = paddle.exp(-paddle.log(paddle.to_tensor(10000.0)) / d_model * div_term)
        div_term1 = paddle.arange(1, d_model, 2, dtype='float32')
        div_term1 = paddle.exp(-paddle.log(paddle.to_tensor(10000.0)) / d_model * div_term1)

        pe[:, 0::2] = paddle.sin(position * div_term)
        pe[:, 1::2] = paddle.cos(position * div_term1)
        pe = pe.unsqueeze(1)
        self.register_buffer("pe", pe, persistable=True)

    def forward(self, x):
        return x + self.pe[: x.shape[0], :]


def masked_fill(x, mask, value):
    y = paddle.full(x.shape, value, x.dtype)
    return paddle.where(mask, y, x)


class Transformer(nn.Layer):
    def __init__(self, input_channels, output_channels=1, time_steps=1, nhead=10, dim_feedforward=512, num_layers=4,
                 dropout=0.1):
        super(Transformer, self).__init__()
        self.model_type = 'Transformer'
        self.time_steps = time_steps
        self.src_mask = None
        self.pos_encoder = PositionalEncoding(input_channels)
        self.encoder_layer = nn.TransformerEncoderLayer(d_model=input_channels, nhead=nhead,
                                                        dim_feedforward=dim_feedforward, dropout=dropout)
        self.transformer_encoder = nn.TransformerEncoder(self.encoder_layer, num_layers=num_layers)

        initrange = 0.1
        self.decoder = nn.Linear(
            input_channels, output_channels,
            weight_attr=paddle.nn.initializer.Uniform(-initrange, initrange),
            bias_attr=paddle.nn.initializer.Constant()
        )

    def _generate_square_subsequent_mask(self, length):
        mask = (paddle.triu(paddle.ones([length, length])) == 1).T.astype('float')
        mask = masked_fill(x=mask, mask=mask == 0, value=float('-inf'))
        mask = masked_fill(x=mask, mask=mask == 1, value=float('0.0'))
        return mask

    def forward(self, src):
        src_len = src.shape[1]
        if self.src_mask is None or self.src_mask.shape[1] != src_len:
            mask = self._generate_square_subsequent_mask(src_len)
            self.src_mask = mask

        src = self.pos_encoder(src)
        output = self.transformer_encoder(src, self.src_mask)
        output = output[:, -self.time_steps:, :]

        output = self.decoder(output)
        return output


# Transformer 输入形状(batch_size, time_steps, features) 输出形状(batch_size, ?)
# paddle.summary(Transformer(15, 4, nhead=3), (12, 187, 15))


class Net(nn.Layer):
    def __init__(self, input_channels=23, output_channels=4, time_steps=1, model='TCN'):
        super(Net, self).__init__()
        if model == 'LSTM':
            self.net = LSTM(input_channels, output_channels)
        elif model == 'TriLSTM':
            self.net = TriLSTM(input_channels, output_channels)
        elif model == 'Transformer':
            self.net = Transformer(input_channels, output_channels, time_steps, nhead=1)
        elif model == 'TCN':
            self.net = TimeSeriesNetwork(input_channels, output_channels, time_steps, num_channels=[256])
        elif model == 'MFTcNeXt':
            self.net = MFTcNeXt(input_channels, output_channels, time_steps, num_channels=[256], hidden_channels=32)

        else:
            raise Exception('Choose Model')

    def forward(self, x):
        y = self.net(x)
        return y



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# @Time    : 2022/5/28 14:11
# @Author  : 马赫
# @Email   : 1692303843@qq.com
# @FileName: olga_func.py


import re
import os
import pandas as pd
import matplotlib.pyplot as plt
try:
    from BiVFM.func import *
except Exception:
    from func import *


def olga_columns_info_parse(olga_columns_type_list):
    """
    OLGA列名信息拆分
    输入列名列表
    返回列名拆分列表

    示例输入
    ["VOLGBL 'GLOBAL' '(-)' 'Global max volume error since last write'",
    "HT 'GLOBAL' '(S)' 'Time step'",
    "GL 'BOUNDARY:' 'BRANCH:' 'PIPELINE' 'PIPE:' 'p2' 'NR:' '2'  '(KG/S)' 'Liquid bulk mass flow'",
    "ID 'BOUNDARY:' 'BRANCH:' 'PIPELINE' 'PIPE:' 'p2' 'NR:' '2'  '(-)' 'Flow regime: 1=Stratified, 2=Annular, 3=Slug, 4=Bubble.'",
    "PT 'SECTION:' 'BRANCH:' 'PIPELINE' 'PIPE:' 'p2' 'NR:' '2'  '(PA)' 'Pressure'",
    "LIQC 'BRANCH:' 'PIPELINE' '(M3)' 'Total liquid content in branch'",
    "CGGBOU 'NODE:' 'OUTLET' '(KG/S)' 'Gas mass flow for each component at boundary node'",
    "GLTHLBOU 'NODE:' 'OUTLET' '(KG/S)' 'Oil mass flow at boundary node'",
    "GLTWTBOU 'NODE:' 'OUTLET' '(KG/S)' 'Water mass flow at boundary node'",
    "GTBOU 'NODE:' 'OUTLET' '(KG/S)' 'Total mass flow at boundary node'",
    "PT 'NODE:' 'OUTLET' '(PA)' 'Pressure'",
    "PTBOU 'NODE:' 'OUTLET' '(PA)' 'Pressure at boundary node'"]


    示例返回

    [['VOLGBL', '', '', 'Global max volume error since last write'],
    ['HT', 'S', '', 'Time step'],
    ['GL', 'KG/S', 'PIPELINE.p2.2', 'Liquid bulk mass flow'],
    ['ID',
    '',
    'PIPELINE.p2.2',
    'Flow regime: 1=Stratified, 2=Annular, 3=Slug, 4=Bubble.'],
    ['PT', 'BAR', 'PIPELINE.p2.2', 'Pressure'],
    ['LIQC', 'M3', 'PIPELINE', 'Total liquid content in branch'],
    ['CGGBOU',
    'KG/S',
    'OUTLET',
    'Gas mass flow for each component at boundary node'],
    ['GLTHLBOU', 'KG/S', 'OUTLET', 'Oil mass flow at boundary node'],
    ['GLTWTBOU', 'KG/S', 'OUTLET', 'Water mass flow at boundary node'],
    ['GTBOU', 'KG/S', 'OUTLET', 'Total mass flow at boundary node'],
    ['PT', 'BAR', 'OUTLET', 'Pressure'],
    ['PTBOU', 'BAR', 'OUTLET', 'Pressure at boundary node']]
    """

    olga_columns_type_split_list = []
    for olga_columns_type_i in olga_columns_type_list:
        _key = re.findall("(.*?) '.*", olga_columns_type_i)[0]
        _describe = re.findall("'(.*?)'", olga_columns_type_i)

        unit = _describe[-2][1:-1]
        if _describe[-2][1:-1] == '-':
            unit = ''

        position = ''
        for _name in _describe[1: -2]:
            if _name in ['BRANCH:', 'PIPE:', 'NR:']:
                if _name in ['PIPE:', 'NR:']:
                    position += '.'
                continue
            position += _name
        describe = _describe[-1]
        # 关键字 [单位] (位置) 解释
        # key = '%s\t[%s]\t(%s)\t"%s"' % (_key, unit, position, describe)
        columns_info = [_key, unit, position, describe]
        olga_columns_type_split_list.append(columns_info)
    return olga_columns_type_split_list


def trend_plot_data(olga_time_plot_path):
    """
    从OLGA时间趋势数据文件中提取时间数据
    olga_time_plot_path: 'xxx.tpl'文件
    """
    with open(olga_time_plot_path) as f:
        str_data = f.readlines()
    str_data = [_.strip() for _ in str_data]

    # 从列名处切分，前面是olga信息，后面是列数据
    olga_info_lines_id = str_data.index("CATALOG")
    # 一共统计多少列变量
    olga_columns_num = int(str_data[olga_info_lines_id + 1])
    olga_columns_type = str_data[olga_info_lines_id + 2:olga_info_lines_id + 2 + olga_columns_num]

    # 将获取到的列名进行切分
    olga_columns_type_dict = olga_columns_info_parse(olga_columns_type)
    olga_data_df = str_data[olga_info_lines_id + 2 + olga_columns_num + 1:]

    olga_data_df = pd.DataFrame([_.split() for _ in olga_data_df], dtype='float')

    # olga_data_df[0] = pd.to_datetime(olga_data_df[0], unit='s')
    olga_data_df.set_index(0, inplace=True)
    olga_data_df.index.name = 'Time(sec)'
    olga_data_df.columns = ['%s[%s](%s)"%s"' % tuple(info) for info in olga_columns_type_dict]

    return olga_data_df


def profile_plot_data(olga_profile_plot_path):
    """
    从OLGA时间剖面数据文件中提取剖面数据
    olga_profile_plot_path: 'xxx.ppl'文件
    """
    with open(olga_profile_plot_path) as f:
        str_data = f.readlines()
    str_data = [_.strip() for _ in str_data]

    # 从列名处切分，前面是olga信息，后面是列数据
    olga_info_lines_id = str_data.index("CATALOG")

    # 获取沿程长度
    olga_info_x_alone_id = str_data[:olga_info_lines_id].index("'PIPELINE'")
    _olga_info_x_alone = str_data[olga_info_x_alone_id + 2: olga_info_x_alone_id + 2 + (
                olga_info_lines_id - olga_info_x_alone_id) // 2 - 1]
    olga_info_x_alone = []
    for _ in _olga_info_x_alone:
        olga_info_x_alone += _.split(' ')
    olga_info_x_alone = [float('%.3g' % float(i)) for i in olga_info_x_alone]

    # 一共统计多少列变量
    olga_columns_num = int(str_data[olga_info_lines_id + 1])
    olga_columns_type = str_data[olga_info_lines_id + 2:olga_info_lines_id + 2 + olga_columns_num]

    # 将获取到的列名进行切分
    olga_columns_type_dict = olga_columns_info_parse(olga_columns_type)
    olga_data_list = str_data[olga_info_lines_id + 2 + olga_columns_num + 1:]
    olga_data_list = [i.split(' ') for i in olga_data_list]

    # 分离出每个时间点的数据
    values_list = []
    values = []
    time_step = float(olga_data_list[0][0])
    for i in olga_data_list:
        if len(i) == 1:
            if len(values) == 0:
                continue
            values_list.append([time_step, values])
            values = []
            time_step = float(i[0])
        else:
            values.append(i)
    values_list.append([time_step, values])

    # 每个时间点都是一个pandas
    """
    注意！！！
    这里如果有的沿程坐标缺少某属性的值，则会把缺失值填pandas.NaN
    """
    olga_data_time_steps = []
    for i in values_list:
        time_step = i[0]
        values = i[1]
        values = pd.DataFrame(values, dtype='float').T
        values.columns = ['%s[%s](%s)"%s"' % tuple(info) for info in olga_columns_type_dict]
        values.index = olga_info_x_alone
        olga_data_time_steps.append([time_step, values])

    return olga_data_time_steps


def olga_trend2csv(olga_time_plot_path, save_path, is_show=False):
    """
    将OLGA生成的tpl文件转为csv文件
    """
    # olga_time_plot_path = 'olga_work1/q4.tpl' # 时间节点信息
    # 趋势数据
    olga_trend_data = trend_plot_data(olga_time_plot_path)

    if is_show:
        for i in range(olga_trend_data.shape[1]):
            olga_trend_data.iloc[:, i].plot(legend=False)
            plt.show()

    # 将趋势数据保存为csv文件
    makedirs(save_path)
    olga_trend_data.to_csv(save_path)


def tpl_list2raw(tpl_dir, raw_csv_dir):
    """
    将tpl文件夹转为csv文件夹
    """
    rm_dir(raw_csv_dir)
    makedirs(raw_csv_dir)

    for i in os.listdir(tpl_dir):
        # 避免`.ipynb_checkpoints`文件
        if not i.endswith('.tpl'):
            continue
        # 将OLGA的.tpl文件转为cvs格式
        _old_path = '%s/%s' % (tpl_dir, i)
        _new_path = '%s/%s.csv' % (raw_csv_dir, os.path.splitext(i)[0])

        olga_trend2csv(olga_time_plot_path=_old_path, save_path=_new_path, is_show=False)


def featrue_extract(path_data, path_save, extract_x, extract_y):
    """
    将OLGA提取出的数据选择其中需要的部分，前几列为x，后面的其余列为y
    extract_x = [
                    [11],  # 保存的列号
                    ['Pressure outlet'],  # 修改后保存的列名
                    [0.00001],  # 单位变化多少倍。 压力单位从Pa变为bar
                    ['PTBOU[PA](OUTLET)"Pressure at boundary node"']  # 修改前的列名
                ]

    extract_y = [
                    [7, 6],
                    ['Oil mass', 'Gas mass'],
                    [1, 1],  # 单位变化多少倍。无变化
                    [
                    'GLTHLBOU[KG/S](OUTLET)"Oil mass flow at boundary node"',
                    'CGGBOU[KG/S](OUTLET)"Gas mass flow for each component at boundary node"'
                    ]
                ]
    featrue_extract(
                    path_data='olga_work1/dataset0/my_1_slug.csv',
                    path_save='olga_work1/dataset1/my_1_slug.csv',
                    extract_x=extract_x, extract_y=extract_y)
    """

    _raw_data = pd.read_csv(path_data, header=0,
                            index_col=0,
                            parse_dates=[0],
                            date_parser=lambda x: df_parser(x, mode=0, unit='s', date_format='%Y-%m-%d %H:%M:%S'),
                            engine='python'
                            )

    # 截取xy参数
    part_df = _raw_data.iloc[:, extract_x[0] + extract_y[0]]
    # 乘单位换算系数
    part_df = part_df * (extract_x[2] + extract_y[2])
    # print('Origin column names:')
    # for i in part_df.columns:
    #     print('\t', i)
    assert part_df.columns.tolist() == extract_x[3] + extract_y[3], \
        '原始列名跟修改后的列名不一致，请检查\n%s\n%s' % (part_df.columns.tolist(), extract_x[3] + extract_y[3])
    part_df.columns = extract_x[1] + extract_y[1]
    # print('Extract column names:')
    # for i in part_df.columns:
    #     print('\t', i)
    part_df.to_csv(path_save)


def featrue_extract_dir(raw_csv_dir, xy_csv_dir, extract_x, extract_y):
    """
    `featrue_extract`函数的文件夹批量模式
    将文件夹1提取指定的xy到文件夹2

    """
    rm_dir(xy_csv_dir)
    makedirs(xy_csv_dir)

    for i in os.listdir(raw_csv_dir):
        # 避免临时缓存文件影响
        if i == '.ipynb_checkpoints':
            continue
        old = '%s/%s' % (raw_csv_dir, i)
        new = '%s/%s' % (xy_csv_dir, i)

        featrue_extract(
            path_data=old,
            path_save=new,
            extract_x=extract_x,
            extract_y=extract_y
        )








#!/usr/bin/python3
# -*- coding: utf-8 -*-
# @Time    : 2022/5/28 11:27
# @Author  : 马赫
# @Email   : 1692303843@qq.com
# @FileName: feature_engineering.py.py
# 特征工程  feature_engineering(raw_data)


def feature_engineering(raw_data):
    """
    特征工程

    输入与输出都是pd.DataFrame

    """

    new_data = raw_data.copy()

    # 构建压力差值作为新特征

    new_data['P_choke_diff_a'] = raw_data['Pressure head_a'] - raw_data['Pressure choke_a']
    new_data['P_choke_ratio_a'] = raw_data['Pressure head_a'] / raw_data['Pressure choke_a']
    new_data['P_choke_sum_diff_a'] = raw_data['Pressure choke_a'] - raw_data['Pressure outlet']

    new_data['P_choke_diff_b'] = raw_data['Pressure head_b'] - raw_data['Pressure choke_b']
    new_data['P_choke_ratio_b'] = raw_data['Pressure head_b'] / raw_data['Pressure choke_b']
    new_data['P_choke_sum_diff_b'] = raw_data['Pressure choke_b'] - raw_data['Pressure outlet']

    new_data['P_choke_diff_c'] = raw_data['Pressure head_c'] - raw_data['Pressure choke_c']
    new_data['P_choke_ratio_c'] = raw_data['Pressure head_c'] / raw_data['Pressure choke_c']
    new_data['P_choke_sum_diff_c'] = raw_data['Pressure choke_c'] - raw_data['Pressure outlet']

    # 构建温度差值作为新特征

    new_data['TM_choke_diff_a'] = raw_data['TM head_a'] - raw_data['TM choke_a']
    new_data['TM_choke_ratio_a'] = raw_data['TM head_a'] / raw_data['TM choke_a']
    new_data['TM_choke_sum_diff_a'] = raw_data['TM choke_a'] - raw_data['TM outlet']

    new_data['TM_choke_diff_b'] = raw_data['TM head_b'] - raw_data['TM choke_b']
    new_data['TM_choke_ratio_b'] = raw_data['TM head_b'] / raw_data['TM choke_b']
    new_data['TM_choke_sum_diff_b'] = raw_data['TM choke_b'] - raw_data['TM outlet']

    new_data['TM_choke_diff_c'] = raw_data['TM head_c'] - raw_data['TM choke_c']
    new_data['TM_choke_ratio_c'] = raw_data['TM head_c'] / raw_data['TM choke_c']
    new_data['TM_choke_sum_diff_c'] = raw_data['TM choke_c'] - raw_data['TM outlet']

    # 构建总流量差值和变化率
    new_data['QT_diff of change'] = raw_data['QT'].diff().fillna(0)

    new_data['QT_rate of change'] = raw_data['QT']
    new_data['QT_rate of change'][1:] = raw_data['QT'][1:] / raw_data['QT'][:-1].values
    new_data['QT_rate of change'][0] = 1

    anew_columns_order = [

        'Pressure head_a',
        'Pressure choke_a',
        'Pressure head_b',
        'Pressure choke_b',
        'Pressure head_c',
        'Pressure choke_c',
        'Pressure outlet',

        # x: P variant
        'P_choke_diff_a',
        'P_choke_ratio_a',
        'P_choke_sum_diff_a',
        'P_choke_diff_b',
        'P_choke_ratio_b',
        'P_choke_sum_diff_b',
        'P_choke_diff_c',
        'P_choke_ratio_c',
        'P_choke_sum_diff_c',

        'TM head_a',
        'TM choke_a',
        'TM head_b',
        'TM choke_b',
        'TM head_c',
        'TM choke_c',
        'TM outlet',

        # x: TM variant
        'TM_choke_diff_a',
        'TM_choke_ratio_a',
        'TM_choke_sum_diff_a',
        'TM_choke_diff_b',
        'TM_choke_ratio_b',
        'TM_choke_sum_diff_b',
        'TM_choke_diff_c',
        'TM_choke_ratio_c',
        'TM_choke_sum_diff_c',

        # x: QT and variant
        'QT_diff of change',
        'QT_rate of change',

        'QT',

        # y
        'QL',
        'QO',
        'QW',
        'QG',
    ]
    new_data = new_data.loc[:, anew_columns_order]

    return new_data



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# @Time    : 2022/6/18 18:42
# @Author  : 马赫
# @Email   : 1692303843@qq.com
# @FileName: visual.py

from BiVFM.train_func import *

import matplotlib.dates as mdates

base_yaml_path = 'olga_work1/conf_0_base.yaml'

base_conf = get_base_config(base_yaml_path)
# print(base_conf)

pkl_dir = base_conf['pkl_dir']
raw_data_list = pickle_load('%s/raw_data_list.pkl' % pkl_dir)


i_data = raw_data_list[0]
#%%

PT_Qowg_data = i_data

index_plt = PT_Qowg_data.index



fig = plt.figure(figsize=(10, 9), dpi=300)
ax1 = fig.add_subplot(2,1,1)

ax1.scatter(index_plt, PT_Qowg_data.loc[:, 'P_Mpa_choke_ahead'], c='#328655', s=1, label='P')
plt.xticks(rotation=25)
plt.legend(loc='upper left', fontsize='large', markerscale=5)
plt.ylabel('Pressure (MPa)')


ax12 = ax1.twinx()
ax12.scatter(index_plt, PT_Qowg_data.loc[:, 'T_C_choke_ahead'], c='#DF6223', s=1, label='T')

plt.xlim([index_plt[0], index_plt[-1]])
# fig.legend(loc=(0.12, 0.8))
plt.legend(loc='upper right', fontsize='large', markerscale=5)
plt.ylabel('Temperature (℃)')

plt.gca().xaxis.set_major_formatter(mdates.DateFormatter('%M:%S'))


ax2 = fig.add_subplot(2,1,2)
plt.xticks([])


# b #203A97
# r #DF6223
# g #328655
# yellow #F4C944



ax2.scatter(index_plt, PT_Qowg_data.loc[:, 'QO'], c='#328655', s=1, label='$Q_o$')
ax2.scatter(index_plt, PT_Qowg_data.loc[:, 'QW'], c='#203A97', s=1, label='$Q_w$')
plt.title('Time')

plt.ylabel('$Q_o, Q_w (m^3/d)$')

plt.legend(loc='upper left', fontsize='large', markerscale=5)

ax21 = ax2.twinx()
ax21.scatter(index_plt, PT_Qowg_data.loc[:, 'QG'], c='#DF6223', s=1, label='$Q_g$')

plt.xlim([index_plt[0], index_plt[-1]])
# fig.legend(loc=(0.09, 0.75))
plt.legend(loc='upper right', fontsize='large', markerscale=5)
plt.ylabel('$Q_g (m^3/d)$')

plt.savefig('draw_PTQ.jpg', dpi=300, bbox_inches='tight')

plt.show()







#!/usr/bin/python3
# -*- coding: utf-8 -*-
# @Time    : 2022/6/7 18:39
# @Author  : 马赫
# @Email   : 1692303843@qq.com
# @FileName: verify.py

from BiVFM.func import *
from BiVFM.olga_func import *
from BiVFM.model import *
from BiVFM.feature_engineering import *
from BiVFM.config_transfer import *
from BiVFM.train_func import *

def valid_metric():
    # 误差指标分析
    print('==================== 性能指标分析 ====================')

    def MAE(y, y_pre):
        return round(np.mean(np.abs(y - y_pre)), 7)

    def MSE(y, y_pre):
        return round(np.mean((y - y_pre) ** 2), 7)

    def RMSE(y, y_pre):
        return round(np.sqrt(MSE(y, y_pre)), 7)

    def MAPE(y, y_pre):
        return round(np.mean(np.abs((y - y_pre) / y)), 7)

    def R2(y, y_pre):
        u = np.sum((y - y_pre) ** 2)
        v = np.sum((y - np.mean(y)) ** 2)
        return round(1 - (u / v), 7)

    def loss_list(y, y_pre):
        print('MAE', MAE(y, y_pre))
        print('MSE', MSE(y, y_pre))
        print('RMSE', RMSE(y, y_pre))
        print('MAPE', MAPE(y, y_pre))
        print('R2', R2(y, y_pre))


    loss_y, loss_y_pred = y_gt[:100, 0], y_predict[:100, 0]
    loss_list(loss_y, loss_y_pred)
    print()
    loss_y, loss_y_pred = y_gt[100: 300, 0], y_predict[100: 300, 0]
    loss_list(loss_y, loss_y_pred)
    print()
    loss_y, loss_y_pred = y_gt[:, 0], y_predict[:, 0]
    loss_list(loss_y, loss_y_pred)






#!/usr/bin/python3
# -*- coding: utf-8 -*-
# @Time    : 2022/6/6 17:07
# @Author  : 马赫
# @Email   : 1692303843@qq.com
# @FileName: olga_extract_tpl.py

import os
import shutil

try:
    from BiVFM.feature_engineering import feature_engineering
except Exception:
    pass

from BiVFM.olga_func import *
from BiVFM.config_transfer import *


def extract_tpl(autoolga_output_dir, tpl_dir):
    """
    extract the **`.tpl`** file from the project directory generated by AutoOLGAFleet
    :param autoolga_output_dir:olga project collection total path
    :param tpl_dir:`.tpl` file extraction path
    :return: None
    """

    try:
        shutil.rmtree(tpl_dir)
    except Exception:
        pass
    os.makedirs(tpl_dir)

    olga_proj_list = os.listdir(autoolga_output_dir)
    for olga_proj in olga_proj_list:
        origin_olga_tpl_filename = '%s/%s/%s.tpl' % (autoolga_output_dir, olga_proj, olga_proj)
        migrate_olga_tpl_filename = '%s/%s.tpl' % (tpl_dir, olga_proj)

        shutil.copy(origin_olga_tpl_filename, migrate_olga_tpl_filename)


def extract_csv(tpl_dir, raw_csv_dir):
    """
    extract the csv file from tpl list


    # 0	PT[PA](FLOWPATH_1.PIPE-1.1)"Pressure"
    # 1	TM[C](FLOWPATH_1.PIPE-1.1)"Fluid temperature"
    # 2	PT[PA](FLOWPATH_1.PIPE-1.3)"Pressure"
    # 3	TM[C](FLOWPATH_1.PIPE-1.3)"Fluid temperature"
    # 4	QG[M3/S](FLOWPATH_1.PIPE-1.6)"Gas volume flow"
    # 5	QLT[M3/S](FLOWPATH_1.PIPE-1.6)"Total liquid volume flow"
    # 6	QLTHL[M3/S](FLOWPATH_1.PIPE-1.6)"Volumetric flow rate oil"
    # 7	QLTWT[M3/S](FLOWPATH_1.PIPE-1.6)"Volumetric flow rate water"
    # 8	QT[M3/S](FLOWPATH_1.PIPE-1.6)"Total volume flow"
    # 9	GT[KG/S](PIPELINE.PIPE-1.2)"Total mass flow"
    # 10	PT[PA](PIPELINE.PIPE-1.2)"Pressure"
    # 11	QT[M3/S](PIPELINE.PIPE-1.2)"Total volume flow"
    # 12	TM[C](PIPELINE.PIPE-1.2)"Fluid temperature"
    # 13	GTSOUR[KG/S](SOURCE-1)"Source mass rate"

    :param tpl_dir:
    :param raw_csv_dir:
    :return:
    """

    try:
        shutil.rmtree(raw_csv_dir)
    except Exception:
        pass
    os.makedirs(raw_csv_dir)

    tpl_list2raw(tpl_dir=tpl_dir, raw_csv_dir=raw_csv_dir)

    _raw_data = pd.read_csv('%s/AUTO_0.csv' % raw_csv_dir, header=0,
                            index_col=0,
                            parse_dates=[0],  # The first column is the time series
                            date_parser=lambda x: df_parser(x, mode=0, unit='s', date_format='%Y-%m-%d %H:%M:%S')
                            )
    for num, i in enumerate(_raw_data.columns.to_list()):
        print('%s\t%s' % (num, i))

    print(_raw_data.shape)


def extract_xy(raw_csv_dir, xy_dir, pkl_dir, extract_xy, resample_time):
    """
    extract xy_dataset from raw_csv
    :param extract_xy: str 'dataset/feature_extractor.xlsx' base_config['extract_xy']
    :param raw_csv_dir:
    :param xy_dir:
    :param resample_time: str Time drop sampling, '2S'
    :return:
    """
    #
    # try:
    #     shutil.rmtree(xy_dir)
    # except Exception:
    #     pass
    # os.makedirs(xy_dir)
    # extract_x = pd.read_excel(extract_xy, sheet_name='x').iloc[:, :-1].dropna().values.T
    # extract_x[0] = extract_x[0].astype('int')
    # extract_x = [list(i) for i in extract_x]
    #
    # extract_y = pd.read_excel(extract_xy, sheet_name='y').iloc[:, :-1].dropna().values.T
    # extract_y[0] = extract_y[0].astype('int')
    # extract_y = [list(i) for i in extract_y]
    #
    # featrue_extract_dir(raw_csv_dir=raw_csv_dir, xy_csv_dir=xy_dir,
    #                     extract_x=extract_x, extract_y=extract_y)

    # read csv data
    csv_order = []
    raw_data_list = []

    for filename in os.listdir(xy_dir):
        if not filename.endswith('.csv'):
            continue
        csv_order.append(filename)

    # In ascending order between underscores and `.xlsx` letters
    csv_order = sorted(csv_order, key=lambda x: int(x[5: -4]))

    csv_order = csv_order[:]

    for filename in csv_order:
        path = os.path.join(xy_dir, filename)
        raw_data = read_df_second_resample(filename=path, resample_time=resample_time)

        # 特征工程变换
        try:

            raw_data_fe = feature_engineering(raw_data)
        except Exception as e:
            raw_data_fe = raw_data

        raw_data_list.append(raw_data_fe)

    # save raw_data_list as pkl file
    try:
        shutil.rmtree(pkl_dir)
    except Exception:
        pass
    os.makedirs(pkl_dir)
    pickle_save(path='%s/raw_data_list.pkl' % pkl_dir,
                content=raw_data_list)

    # raw_data_list = pickle_load('%s/raw_data_list.pkl' % pkl_dir)

    print('Number of sample files:', len(raw_data_list))


if __name__ == '__main__':

    step = 3
    base_yaml_path = 'olga_work1/conf_0_base.yaml'
    # base_yaml_path = r'D:\work\VFM\BiVFM\experiment\e2_multiwell\olga_work2\conf_0_base.yaml'

    base_conf = get_base_config(base_yaml_path)
    # print(base_conf)

    # Step 1: extract the **`.tpl`** file from the project directory generated by AutoOLGAFleet
    if step == 1:
        extract_tpl(base_conf['autoolga_output_dir'], base_conf['tpl_dir'])

    # Step 2: extract csv from tpl
    if step == 2:
        extract_csv(base_conf['tpl_dir'], base_conf['raw_csv_dir'])

    # Step 3: extract xy from csv and then form pkl
    if step == 3:
        extract_xy(base_conf['raw_csv_dir'], base_conf['xy_dir'], base_conf['pkl_dir'],
                   base_conf['extract_xy'],
                   base_conf['resample_time'])



import codecs
import os
import sys

import setuptools

"""
打包的用的setup必须引入，
"""

def read(fname):
    """
    定义一个read方法，用来读取目录下的长描述
    我们一般是将README文件中的内容读取出来作为长描述，这个会在PyPI中你这个包的页面上展现出来，
    你也可以不用这个方法，自己手动写内容即可，
    PyPI上支持.rst格式的文件。暂不支持.md格式的文件，<BR>.rst文件PyPI会自动把它转为HTML形式显示在你包的信息页面上。
    """
    return codecs.open(os.path.join(os.path.dirname(__file__), fname)).read()
setuptools.setup(
  name="HVFM",
  version="1.0.1",
  author="He Ma",
  author_email="mahepetroleum@163.com",
  description="Paper: Hybrid Virtual Flow Metering on Arbitrary Well Patterns for Transient Multiphase Prediction Driven by Mechanistic and Data Model",
  long_description=read("README.md"),
  long_description_content_type="text/markdown",
  url="https://pypi.org/project/BiVFM-Fleet",
  packages=setuptools.find_packages(),
  classifiers=[
  "Programming Language :: Python :: 3",
  "License :: OSI Approved :: MIT License",
  "Operating System :: OS Independent",
  ],
  keywords="Virtual flow meter, neural network",

)


#!/usr/bin/python3
# -*- coding: utf-8 -*-
# @Time    : 2022/6/7 18:39
# @Author  : 马赫
# @Email   : 1692303843@qq.com
# @FileName: verify.py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from BiVFM.func import *
from BiVFM.olga_func import *
from BiVFM.model import *
from BiVFM.feature_engineering import *
from BiVFM.config_transfer import *
from BiVFM.train_func import *

# %% BASE LOAD
base_yaml_path = 'olga_work1/conf_1_MFTcNeXt_epoch300.yaml'

base_conf = get_base_config(base_yaml_path)
# print(base_conf)

pkl_dir = base_conf['pkl_dir']
num_epoch = base_conf['num_epoch']
input_channels = base_conf['input_channels']
output_channels = base_conf['output_channels']

continue_epoch = base_conf['continue_epoch']  # 从多少轮开始训练

ratio = base_conf['ratio']
model_name = base_conf['model_name']

load_model = base_conf['load_model']

# 加载数据
raw_data_list = pickle_load('%s/raw_data_list.pkl' % pkl_dir)
print('Number of sample files:', len(raw_data_list))

# 以全部文件作为标准化基准
all_mean = []
all_std = []
for i in raw_data_list:
    mean = i.mean()
    std = i.std()
    all_mean.append(mean)
    all_std.append(std)

all_mean = pd.concat(all_mean, axis=1).T
all_std = pd.concat(all_std, axis=1).T

standardized_args = (all_mean.mean(), all_std.mean())

stand_data_list = [raw2stand(raw_data, standardized_args) for raw_data in raw_data_list]

feature_list = np.array(raw_data_list[0].columns)

x_feature = list(range(len(feature_list)))[:-output_channels]
y_feature = list(range(len(feature_list)))[-output_channels:]

print('x特征%s列，y特征%s列' % (len(x_feature), len(y_feature)))
print('========== x_feature: ==========\n%s' % feature_list[x_feature])
print('========== y_feature: ==========\n%s' % feature_list[y_feature])

np.random.seed(0)
np.random.shuffle(stand_data_list)

num_ratio = int(len(stand_data_list) * ratio)
train_data_list = stand_data_list[: num_ratio]
test_data_list = stand_data_list[num_ratio:]

# DataLoader
train_reader = MyIterableDataset(train_data_list, base_conf['x_length'], base_conf['y_length'], x_feature,
                                 y_feature)

test_reader = MyIterableDataset(test_data_list[: 10], base_conf['x_length'], base_conf['y_length'],
                                x_feature,
                                y_feature, return_y_info=False)

train_loader = DataLoader(train_reader, batch_size=base_conf['batch_size'], drop_last=False)
test_loader = DataLoader(test_reader, batch_size=base_conf['batch_size'], drop_last=False)

for batch_id, (data, label) in enumerate(train_loader()):
    print('train batch x, y_gt shape\t', data.shape, label.shape)
    break

# scalar_s2r = None
# scalar_s2r_x = (standardized_args[0].iloc[x_feature].values, standardized_args[1].iloc[x_feature].values)
scalar_s2r_y = (standardized_args[0].iloc[y_feature].values, standardized_args[1].iloc[y_feature].values)

# %% model list verify

verify_ans = {'model_name': [], 'y_predict': [], 'y_gt': []}

base_yaml_path_list = [
    'olga_work1/conf_1_LSTM_epoch300.yaml',
    'olga_work1/conf_1_Transformer_epoch300.yaml',
    'olga_work1/conf_1_TCN_epoch300.yaml',
    'olga_work1/conf_1_MFTcNeXt_epoch300.yaml',
]
for base_yaml_path in base_yaml_path_list:
    base_conf = get_base_config(base_yaml_path)
    # print(base_conf)

    pkl_dir = base_conf['pkl_dir']
    num_epoch = base_conf['num_epoch']
    input_channels = base_conf['input_channels']
    output_channels = base_conf['output_channels']

    continue_epoch = base_conf['continue_epoch']  # 从多少轮开始训练

    ratio = base_conf['ratio']
    model_name = base_conf['model_name']

    load_model = base_conf['load_model']


    a = Net(input_channels, output_channels, time_steps=base_conf['y_length'], model=model_name)

    shape = (128, base_conf['y_length'], input_channels)

    print('======= Model:%s, input shape %s, output shape %s =======' % (model_name, shape, a(paddle.randn(shape)).shape))
    # print(paddle.summary(a, shape))
    total_params = paddle.summary(a, shape)['total_params']

    net = Net(input_channels, output_channels, time_steps=base_conf['y_length'], model=model_name)
    opt = Adam(learning_rate=2e-4, parameters=net.parameters())  # Adam
    loss_fc = LossFC()

    if load_model:
        try:
            load_params(net, model_path=base_conf['model_path'], opt=opt, opt_path=base_conf['opt_path'])
            log('模型已加载', log_file=base_conf['other_log_file'])
        except Exception as e:
            log('未加载模型', log_file=base_conf['other_log_file'])

    else:
        log('未加载模型', log_file=base_conf['other_log_file'])


    valid_img_save_path = '%s/log_valid/%s.jpg' % (base_conf['base_dir'], model_name)

    y_predict, y_gt = visualization(model=net, loader=test_loader,
                                    standardized_args=scalar_s2r_y,
                                    is_show=True, save_path=valid_img_save_path,
                                    title=feature_list[y_feature],
                                    show_error=False)

    verify_ans['model_name'].append(model_name)

    verify_ans['y_predict'].append(y_predict)
    verify_ans['y_gt'].append(y_gt)


# %%

def MSE(actual, predict):
    return np.mean((actual - predict) ** 2, axis=0)

def RMSE(actual, predict):
    return np.sqrt(MSE(actual, predict))


def MAE(actual, predict):
    return np.mean(np.abs(actual - predict), axis=0)

def MAPE(actual, predict):
    return np.mean(np.abs((actual - predict) / actual), axis=0)

def MAPE_percentage(actual, predict):
    return np.abs((actual - predict) / actual)


def R2(actual, predict):
    u = np.sum((actual - predict) ** 2, axis=0)
    v = np.sum((actual - np.mean(actual, axis=0)) ** 2, axis=0)
    return 1 - (u / v)


all_metric_df = []
for num, model_name in enumerate(verify_ans['model_name']):
    y_gt, y_predict = verify_ans['y_gt'][num], verify_ans['y_predict'][num]
    rmse = RMSE(y_gt, y_predict)
    mae = MAE(y_gt, y_predict)
    mape = MAPE(y_gt, y_predict)

    metric_df = pd.DataFrame({'rmse': rmse, 'mae': mae, 'mape': mape, }).T
    metric_df.columns = [i+'_'+model_name for i in feature_list[y_feature]]
    all_metric_df.append(metric_df)

all_metric_df = pd.concat(all_metric_df, axis=1)
metric_save_path = '%s/log_valid/metric.csv' % base_conf['base_dir']
all_metric_df.to_csv(metric_save_path)


#%%

#
# print('MSE', MSE(y_gt, y_predict))
# print('RMSE', RMSE(y_gt, y_predict))
#
# print('MAE', MAE(y_gt, y_predict))
# print('MAPE', MAPE(y_gt, y_predict))
# print('R2', R2(y_gt, y_predict))

# mape_perc = MAPE_percentage(y_gt, y_predict)





# %%
actual, predict = y_gt, y_predict
a = np.mean(np.abs((actual - predict) / actual), axis=1)
b = np.percentile(a,50)

#%%

#
# plt.figure(figsize=(10, 5))  # 设置画布的尺寸
# plt.title('Examples of boxplot', fontsize=20)  # 标题，并设定字号大小
#
# # boxprops：color箱体边框色，facecolor箱体填充色；
# plt.boxplot([box_1, box_2, box_3, box_4], patch_artist=True, boxprops={'color': 'orangered', 'facecolor': 'pink'})
#
# plt.show()  # 显示图像


